#
# PST lib Input - special class for input chewing
# 
# (c) 2002  Tomas Valla   <valla@atrey.karlin.mff.cuni.cz>
# 
# This software may be used and distributed under the terms
# of the GNU General Public License. See COPYING for details.
#
#

package Input;
use OM;
@ISA = qw(OM);
use DSC;
use IO::File;
use strict;



sub new {
	my $proto = shift;
	my $input = shift;
	my $output = shift;
	my $class = ref($proto) || $proto;
	my $self = OM->new;

	$self->name("Input");
	$self->{FILE_HANDLER} = DSC->new;    # source file DSC
	$self->{GET_SIZE_MODE} = 0;       # how Boundingbox is computed
	$self->{INPUT_NAME} = $input;
	$self->{OUTPUT_NAME} = $output;

	return bless($self,$class);
}

sub parse_options {
	my $self = shift;
	my $str;

	while ($str = shift) {
		if ($str =~ /^\s*(gsm|get_size_mode)\s*\=\s*(\d+)\s*$/) {
			$self->die("invalid input mode") if $2 > 2;
			$self->{GET_SIZE_MODE} = $2;
		} else { $self->die("bad options\n") }
	}
}

# as a last thing we write document epilog
sub done {
	my $self = shift;
	$self->{FILE_HANDLER}->write_epilog;
	if ($self->{GET_SIZE_MODE} == 2) {
		$self->log(3,"deleting temporary pswrite output");
		unlink($self->{INPUT_NAME} . ".tmp");
	}
}

# computed boundingbox by scannig DSC comments
# fast but sometimes DSC is wrong...
sub get_size_from_ps {
	my $self = shift;
	my $f = $self->{FILE_HANDLER};
	my $width;
	my $height;
	my $hoff;
	my $voff;


	$self->log(5, "get_size_from_ps start\n");
	$self->{NUM_PAGES} = $f->{NUM_PAGES};
	$self->die("I need BoundingBox\n") if not defined $f->{GLOBALS}{"BoundingBox:"};
	if ($f->{GLOBALS}{"BoundingBox:"} =~ /^\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*$/) {
		$width = $3 - $1;        #      llx    lly     urx     ury
		$height = $4 - $2;
		$hoff = $1;
		$voff = $2;
	}else { $self->die("invalid BoundingBox format"); }
	foreach my $p ( keys %{$f->{PAGE_NUMBERS}} ) {
		$self->{PAGE_WIDTH}{$p} = $width;
		$self->{PAGE_HEIGHT}{$p} = $height;
		$self->{PAGE_H_OFF}{$p} = $hoff;
		$self->{PAGE_V_OFF}{$p} = $voff;
		my $pbb = $f->{PAGE_ATTRS}{$p}{"PageBoundingBox:"};
		if(defined $pbb && $pbb =~ /^(\d+)\s+(\d+)\s+(\d+)\s+(\d+)$/) {
		   	$self->{PAGE_WIDTH}{$p} = $3 - $1;
			$self->{PAGE_HEIGHT}{$p} = $4 - $2;
			$self->{PAGE_H_OFF}{$p} = $1;
			$self->{PAGE_V_OFF}{$p} = $2;
		}
	}
}

sub bbox_compute {
	my $self = shift;
	my $input = shift;
	my $f = $self->{FILE_HANDLER};
	my $llx;
	my $lly;
	my $urx;
	my $ury;

	$self->log(4,"now munching GS bbox output...");
	while(<$input>) {
		if (/^%%BoundingBox:\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/) {
			$llx = $1 if not defined $llx or $1<$llx;
			$lly = $2 if not defined $lly or $2<$lly;
			$urx = $3 if not defined $urx or $3>$urx;
			$ury = $4 if not defined $ury or $4>$ury;
		}
	}
	$self->die("Ghostscript did not report any boundingbox")
		unless defined $llx and defined $lly and defined $urx and defined $ury;

	$f->{GLOBALS}{"BoundingBox:"} = "$llx $lly $urx $ury";
	$self->log(2,"Computed Bounddingbox is $llx $lly $urx $ury");
	$self->{NUM_PAGES} = $f->{NUM_PAGES};

	foreach my $p ( keys %{$f->{PAGE_NUMBERS}} ) {
		$self->{PAGE_WIDTH}{$p} = $urx - $llx;
		$self->{PAGE_HEIGHT}{$p} = $ury - $lly;
		$self->{PAGE_H_OFF}{$p} = $llx;
		$self->{PAGE_V_OFF}{$p} = $lly;
	}
}


sub init {
	my $self = shift;
	my $f = $self->{FILE_HANDLER};

	$self->{FILE_HANDLER}->verbose($self->{VERBOSE});

	if ($self->{GET_SIZE_MODE} == 0) {
		$self->log(2,"Relying on BoundingBox from header");
		$f->open($self->{INPUT_NAME}) or $self->die("Cannot open ",$self->{INPUT_NAME});
		my $r = $f->parse;
		$self->die($r) if $r ne "";
		my $fd = new IO::File;
		$fd->open(">".$self->{OUTPUT_NAME}) 
			or $self->die("Cannot open ",$self->{OUTPUT_NAME}, "for writing");
		$f->write_start($fd);
		$self->{OUTPUT} = $fd;

		$self->get_size_from_ps;
	} elsif ($self->{GET_SIZE_MODE} == 1) {
		my $t = new IO::File;
		$f->open($self->{INPUT_NAME}) or $self->die("Cannot open ",$self->{INPUT_NAME});
		my $r = $f->parse;
		$self->die($r) if $r ne "";
		$self->log(2,"Executing ghostscript to compute BoundingBox...");
		$t->open("gs -dNOPAUSE -dBATCH -sDEVICE=bbox " . $self->{INPUT_NAME} . " 2>&1 |")
			or $self->die("Unable to establish pipe with ghostscript");
		
		$self->bbox_compute($t);
		my $fd = new IO::File;
		$fd->open(">".$self->{OUTPUT_NAME}) 
			or $self->die("Cannot open ",$self->{OUTPUT_NAME}, "for writing");
		$f->write_start($fd);
		$self->{OUTPUT} = $fd;
	} elsif ($self->{GET_SIZE_MODE} == 2) {

		$self->log(2,"Executing ghostricpt to completely rewrite input...");
		my $r = 0xffff & system("gs -dNOPAUSE -dBATCH -sDEVICE=pswrite ".
			"-sOutputFile=".$self->{INPUT_NAME}.".tmp ".$self->{INPUT_NAME});

		$self->die("error running ghostscript") unless $r == 0;

		$self->log(3,"Now opening the temporary rewritten file");
		$f->open($self->{INPUT_NAME}.".tmp") or $self->die("Cannot open ",$self->{INPUT_NAME},".tmp");

		$r = $f->parse;
		$self->die($r) if $r ne "";
		my $fd = new IO::File;
		$fd->open(">".$self->{OUTPUT_NAME}) 
			or $self->die("Cannot open ",$self->{OUTPUT_NAME}, "for writing");
		$f->write_start($fd);
		$self->{OUTPUT} = $fd;

		$self->get_size_from_ps;
	} else {
		$self->die("unknown get_size_mode\n");
	}
}


# just dump document header and prolog
sub run {
	my $self = shift;
	$self->log(4,"writing document header and prolog");
	$self->{FILE_HANDLER}->write_header;
	$self->{FILE_HANDLER}->write_prolog;
}

# write the actual body of page
sub write_page {
	my $self = shift;
	my $page = shift;
	my $f = $self->{OUTPUT};

	$self->log(4,"writing page body $page");
	if ($self->{PAGE_H_OFF}{$page} != 0  or $self->{PAGE_V_OFF}{$page} != 0) {
		print $f "userdict/PSTdict save put\nPSTmatrix setmatrix\n",
			-$self->{PAGE_H_OFF}{$page}," ",-$self->{PAGE_V_OFF}{$page}," translate\n",
	                "userdict/PSTmatrix matrix currentmatrix put\n",
	                #"userdict/PSTclip{0 0 moveto\n",
	                #$pw," 0 rlineto 0 ",$ph," rlineto -",$pw," 0 rlineto\n",
	                #"closepath}put initclip\n",
	                "PSTxform concat\n";

		$self->{FILE_HANDLER}->write_page_body($page);
		print $f "PSTdict restore\n";
	} else {
		$self->{FILE_HANDLER}->write_page_body($page);
	}
}

1;  # no init 
